@@ -1,21 +1,20 @@
-import contextlib
 import functools
 import json
 import os.path
 from typing import Any
 from typing import Dict
+from typing import FrozenSet
 from typing import List
 from typing import Match
 from typing import NamedTuple
 from typing import Optional
-from typing import Sequence
 from typing import Tuple
 from typing import TypeVar
 
 from identify.identify import tags_from_filename
 
 from babi._types import Protocol
-from babi.fdict import FDict
+from babi.fdict import FChainMap
 from babi.reg import _Reg
 from babi.reg import _RegSet
 from babi.reg import ERR_REG
@@ -68,6 +67,8 @@ class _Rule(Protocol):
     def include(self) -> Optional[str]: ...
     @property
     def patterns(self) -> 'Tuple[_Rule, ...]': ...
+    @property
+    def repository(self) -> 'FChainMap[str, _Rule]': ...
 
 
 @uniquely_constructed
@@ -84,9 +85,24 @@ class Rule(NamedTuple):
     while_captures: Captures
     include: Optional[str]
     patterns: Tuple[_Rule, ...]
+    repository: FChainMap[str, _Rule]
 
     @classmethod
-    def from_dct(cls, dct: Dict[str, Any]) -> _Rule:
+    def make(
+            cls,
+            dct: Dict[str, Any],
+            parent_repository: FChainMap[str, _Rule],
+    ) -> _Rule:
+        if 'repository' in dct:
+            # this looks odd, but it's so we can have a self-referential
+            # immutable-after-construction chain map
+            repository_dct: Dict[str, _Rule] = {}
+            repository = FChainMap(parent_repository, repository_dct)
+            for k, sub_dct in dct['repository'].items():
+                repository_dct[k] = Rule.make(sub_dct, repository)
+        else:
+            repository = parent_repository
+
         name = _split_name(dct.get('name'))
         match = dct.get('match')
         begin = dct.get('begin')
@@ -96,7 +112,7 @@ class Rule(NamedTuple):
 
         if 'captures' in dct:
             captures = tuple(
-                (int(k), Rule.from_dct(v))
+                (int(k), Rule.make(v, repository))
                 for k, v in dct['captures'].items()
             )
         else:
@@ -104,7 +120,7 @@ class Rule(NamedTuple):
 
         if 'beginCaptures' in dct:
             begin_captures = tuple(
-                (int(k), Rule.from_dct(v))
+                (int(k), Rule.make(v, repository))
                 for k, v in dct['beginCaptures'].items()
             )
         else:
@@ -112,7 +128,7 @@ class Rule(NamedTuple):
 
         if 'endCaptures' in dct:
             end_captures = tuple(
-                (int(k), Rule.from_dct(v))
+                (int(k), Rule.make(v, repository))
                 for k, v in dct['endCaptures'].items()
             )
         else:
@@ -120,7 +136,7 @@ class Rule(NamedTuple):
 
         if 'whileCaptures' in dct:
             while_captures = tuple(
-                (int(k), Rule.from_dct(v))
+                (int(k), Rule.make(v, repository))
                 for k, v in dct['whileCaptures'].items()
             )
         else:
@@ -142,7 +158,7 @@ class Rule(NamedTuple):
         include = dct.get('include')
 
         if 'patterns' in dct:
-            patterns = tuple(Rule.from_dct(d) for d in dct['patterns'])
+            patterns = tuple(Rule.make(d, repository) for d in dct['patterns'])
         else:
             patterns = ()
 
@@ -159,29 +175,33 @@ class Rule(NamedTuple):
             while_captures=while_captures,
             include=include,
             patterns=patterns,
+            repository=repository,
         )
 
 
 @uniquely_constructed
 class Grammar(NamedTuple):
     scope_name: str
+    repository: FChainMap[str, _Rule]
     patterns: Tuple[_Rule, ...]
-    repository: FDict[str, _Rule]
 
     @classmethod
-    def from_data(cls, data: Dict[str, Any]) -> 'Grammar':
+    def make(cls, data: Dict[str, Any]) -> 'Grammar':
         scope_name = data['scopeName']
-        patterns = tuple(Rule.from_dct(dct) for dct in data['patterns'])
         if 'repository' in data:
-            repository = FDict({
-                k: Rule.from_dct(dct) for k, dct in data['repository'].items()
-            })
+            # this looks odd, but it's so we can have a self-referential
+            # immutable-after-construction chain map
+            repository_dct: Dict[str, _Rule] = {}
+            repository = FChainMap(repository_dct)
+            for k, dct in data['repository'].items():
+                repository_dct[k] = Rule.make(dct, repository)
         else:
-            repository = FDict({})
+            repository = FChainMap()
+        patterns = tuple(Rule.make(d, repository) for d in data['patterns'])
         return cls(
             scope_name=scope_name,
-            patterns=patterns,
             repository=repository,
+            patterns=patterns,
         )
 
 
@@ -531,22 +551,23 @@ class Compiler:
     def _include(
             self,
             grammar: Grammar,
+            repository: FChainMap[str, _Rule],
             s: str,
     ) -> Tuple[List[str], Tuple[_Rule, ...]]:
         if s == '$self':
             return self._patterns(grammar, grammar.patterns)
         elif s == '$base':
             grammar = self._grammars.grammar_for_scope(self._root_scope)
-            return self._include(grammar, '$self')
+            return self._include(grammar, grammar.repository, '$self')
         elif s.startswith('#'):
-            return self._patterns(grammar, (grammar.repository[s[1:]],))
+            return self._patterns(grammar, (repository[s[1:]],))
         elif '#' not in s:
             grammar = self._grammars.grammar_for_scope(s)
-            return self._include(grammar, '$self')
+            return self._include(grammar, grammar.repository, '$self')
         else:
             scope, _, s = s.partition('#')
             grammar = self._grammars.grammar_for_scope(scope)
-            return self._include(grammar, f'#{s}')
+            return self._include(grammar, grammar.repository, f'#{s}')
 
     @functools.lru_cache(maxsize=None)
     def _patterns(
@@ -558,7 +579,9 @@ class Compiler:
         ret_rules: List[_Rule] = []
         for rule in rules:
             if rule.include is not None:
-                tmp_regs, tmp_rules = self._include(grammar, rule.include)
+                tmp_regs, tmp_rules = self._include(
+                    grammar, rule.repository, rule.include,
+                )
                 ret_regs.extend(tmp_regs)
                 ret_rules.extend(tmp_rules)
             elif rule.match is None and rule.begin is None and rule.patterns:
@@ -618,8 +641,10 @@ class Compiler:
             return PatternRule(rule.name, make_regset(*regs), rules)
 
     def compile_rule(self, rule: _Rule) -> CompiledRule:
-        with contextlib.suppress(KeyError):
+        try:
             return self._c_rules[rule]
+        except KeyError:
+            pass
 
         grammar = self._rule_to_grammar[rule]
         ret = self._c_rules[rule] = self._compile_rule(grammar, rule)
@@ -627,41 +652,58 @@ class Compiler:
 
 
 class Grammars:
-    def __init__(self, grammars: Sequence[Dict[str, Any]]) -> None:
-        self._raw = {grammar['scopeName']: grammar for grammar in grammars}
-        self._find_scope = [
-            (
-                frozenset(grammar.get('fileTypes', ())),
-                make_reg(grammar.get('firstLineMatch', '$impossible^')),
-                grammar['scopeName'],
-            )
-            for grammar in grammars
-        ]
-        self._parsed: Dict[str, Grammar] = {}
-        self._compilers: Dict[str, Compiler] = {}
+    def __init__(self, *directories: str) -> None:
+        self._scope_to_files = {
+            os.path.splitext(filename)[0]: os.path.join(directory, filename)
+            for directory in directories
+            if os.path.exists(directory)
+            for filename in sorted(os.listdir(directory))
+            if filename.endswith('.json')
+        }
 
-    @classmethod
-    def from_syntax_dir(cls, syntax_dir: str) -> 'Grammars':
-        grammars = [{'scopeName': 'source.unknown', 'patterns': []}]
-        if os.path.exists(syntax_dir):
-            for filename in os.listdir(syntax_dir):
-                with open(os.path.join(syntax_dir, filename)) as f:
-                    grammars.append(json.load(f))
-        return cls(grammars)
+        unknown_grammar = {'scopeName': 'source.unknown', 'patterns': []}
+        self._raw = {'source.unknown': unknown_grammar}
+        self._file_types: List[Tuple[FrozenSet[str], str]] = []
+        self._first_line: List[Tuple[_Reg, str]] = []
+        self._parsed: Dict[str, Grammar] = {}
+        self._compiled: Dict[str, Compiler] = {}
+
+    def _raw_for_scope(self, scope: str) -> Dict[str, Any]:
+        try:
+            return self._raw[scope]
+        except KeyError:
+            pass
+
+        grammar_path = self._scope_to_files.pop(scope)
+        with open(grammar_path) as f:
+            ret = self._raw[scope] = json.load(f)
+
+        file_types = frozenset(ret.get('fileTypes', ()))
+        first_line = make_reg(ret.get('firstLineMatch', '$impossible^'))
+
+        self._file_types.append((file_types, scope))
+        self._first_line.append((first_line, scope))
+
+        return ret
 
     def grammar_for_scope(self, scope: str) -> Grammar:
-        with contextlib.suppress(KeyError):
+        try:
             return self._parsed[scope]
+        except KeyError:
+            pass
 
-        ret = self._parsed[scope] = Grammar.from_data(self._raw[scope])
+        raw = self._raw_for_scope(scope)
+        ret = self._parsed[scope] = Grammar.make(raw)
         return ret
 
     def compiler_for_scope(self, scope: str) -> Compiler:
-        with contextlib.suppress(KeyError):
-            return self._compilers[scope]
+        try:
+            return self._compiled[scope]
+        except KeyError:
+            pass
 
         grammar = self.grammar_for_scope(scope)
-        ret = self._compilers[scope] = Compiler(grammar, self)
+        ret = self._compiled[scope] = Compiler(grammar, self)
         return ret
 
     def blank_compiler(self) -> Compiler:
@@ -669,20 +711,26 @@ class Grammars:
 
     def compiler_for_file(self, filename: str, first_line: str) -> Compiler:
         for tag in tags_from_filename(filename) - {'text'}:
-            with contextlib.suppress(KeyError):
+            try:
+                # TODO: this doesn't always match even if we detect it
                 return self.compiler_for_scope(f'source.{tag}')
+            except KeyError:
+                pass
+
+        # didn't find it in the fast path, need to read all the json
+        for k in tuple(self._scope_to_files):
+            self._raw_for_scope(k)
 
         _, _, ext = os.path.basename(filename).rpartition('.')
-        for extensions, first_line_match, scope_name in self._find_scope:
-            if (
-                    ext in extensions or
-                    first_line_match.match(
-                        first_line, 0, first_line=True, boundary=True,
-                    )
-            ):
-                return self.compiler_for_scope(scope_name)
-        else:
-            return self.compiler_for_scope('source.unknown')
+        for extensions, scope in self._file_types:
+            if ext in extensions:
+                return self.compiler_for_scope(scope)
+
+        for reg, scope in self._first_line:
+            if reg.match(first_line, 0, first_line=True, boundary=True):
+                return self.compiler_for_scope(scope)
+
+        return self.compiler_for_scope('source.unknown')
 
 
 def highlight_line(
